unit DFileStreams;

interface

uses
  Classes;

type
  TDDFileOpenMode = (
    dfBaseCreate,
    dfBaseAlways,
    dfBaseReadRight,
    dfBaseWriteRight,
    dfBaseDenyReadRight,
    dfBaseDenyWriteRight,
    dfBaseDoNotBuffer,
    dfBaseDelete,
    dfBaseHintRandom,
    dfBaseHintSequential
  );
  TDDFileOpenModes = set of TDDFileOpenMode;

  TDFileStream = class(THandleStream)
  public
    ChangeNotify: Boolean;

    constructor Create(const FileName: String; Mode: TDDFileOpenModes);
    destructor Destroy; override;
    function WriteStr(S: String): LongInt;
    function ReadStr(MaxLen: LongInt = MAXINT): String;
    function ReadRIFFChunk(var Typ: LongWord): String;
  protected
    Path: String;
  end;

const
  dfCreate         : TDDFileOpenModes = [dfBaseCreate];
  dfAlways         : TDDFileOpenModes = [dfBaseAlways];
  dfOpenOrFail     : TDDFileOpenModes = [];
  dfOpenOrCreate   : TDDFileOpenModes = [dfBaseAlways];
  dfCreateOrFail   : TDDFileOpenModes = [dfBaseCreate];
  dfCreateOrReplace: TDDFileOpenModes = [dfBaseCreate, dfBaseAlways];

  dfRead     : TDDFileOpenModes = [dfBaseReadRight];
  dfWrite    : TDDFileOpenModes = [dfBaseWriteRight];
  dfReadWrite: TDDFileOpenModes = [dfBaseReadRight, dfBaseWriteRight];

  dfDenyRead : TDDFileOpenModes = [dfBaseDenyReadRight];
  dfDenyWrite: TDDFileOpenModes = [dfBaseDenyWriteRight];
  dfExclusive: TDDFileOpenModes = [dfBaseDenyReadRight, dfBaseDenyWriteRight];
  dfDenyNone : TDDFileOpenModes = [];

  dfNoBuffer : TDDFileOpenModes = [dfBaseDoNotBuffer];
  dfDelete   : TDDFileOpenModes = [dfBaseDelete];

  dfHintRandom    : TDDFileOpenModes = [dfBaseHintRandom];
  dfHintSequential: TDDFileOpenModes = [dfBaseHintSequential];

implementation

uses
  Consts, Windows, ShlObj;

constructor TDFileStream.Create(const FileName: String; Mode: TDDFileOpenModes);
var
  DesiredAccess,
  ShareMode,
  Creation,
  Flags: LongWord;
  NewHandle: THandle;

begin
  Path := FileName;
  DesiredAccess := 0;
  if dfBaseReadRight in Mode then DesiredAccess := DesiredAccess or GENERIC_READ;
  if dfBaseWriteRight in Mode then begin
    DesiredAccess := DesiredAccess or GENERIC_WRITE;
    ChangeNotify := True;
  end
  else
    ChangeNotify := False
  ;

  ShareMode := FILE_SHARE_READ or FILE_SHARE_WRITE;
  if dfBaseDenyReadRight in Mode then ShareMode := ShareMode and not FILE_SHARE_READ;
  if dfBaseDenyWriteRight in Mode then ShareMode := ShareMode and not FILE_SHARE_WRITE;
  if dfBaseDelete in Mode then ShareMode := ShareMode or FILE_SHARE_DELETE;

  if dfBaseCreate in Mode then begin
    if dfBaseAlways in Mode then
      Creation := CREATE_ALWAYS
    else
      Creation := CREATE_NEW
    ;
  end
  else begin
    if dfBaseAlways in Mode then
      Creation := OPEN_ALWAYS
    else
      Creation := OPEN_EXISTING
    ;
  end;

  Flags := FILE_ATTRIBUTE_NORMAL;
  if dfBaseDoNotBuffer in Mode then Flags := Flags or FILE_FLAG_NO_BUFFERING;
  if dfBaseDelete in Mode then Flags := Flags or FILE_FLAG_DELETE_ON_CLOSE;

  if dfBaseHintRandom in Mode then Flags := Flags or FILE_FLAG_RANDOM_ACCESS;
  if dfBaseHintSequential in Mode then Flags := Flags or FILE_FLAG_SEQUENTIAL_SCAN;

  ChangeNotify := (Mode * [dfBaseWriteRight, dfBaseDelete]) <> [];

  NewHandle := CreateFile(
    PChar(FileName),
    DesiredAccess,
    ShareMode,
    nil,
    Creation,
    Flags,
    0
  );

  if NewHandle = 0 then NewHandle := INVALID_HANDLE_VALUE;
  if NewHandle = INVALID_HANDLE_VALUE then begin
    if dfBaseCreate in Mode then
      raise EFOpenError.Create('Cannot create file ' + FileName)
    else
      raise EFOpenError.Create('Cannot open file ' + FileName)
    ;
  end
  else
    inherited Create(NewHandle)
  ;
end;

destructor TDFileStream.Destroy;
begin
  if THandle(Handle) = 0 then Exit;
  if THandle(Handle) = INVALID_HANDLE_VALUE then Exit;
  CloseHandle(Handle);
  
  if ChangeNotify then
    SHChangeNotify(SHCNE_UPDATEITEM, SHCNF_PATH, PChar(Path), nil)
  ;
end;

function TDFileStream.ReadStr(MaxLen: LongInt = MAXINT): String;
var
  Left: Integer;

begin
  Left := Size - Position;
  if MaxLen > Left then
    MaxLen := Left
  ;
  SetString(Result, nil, MaxLen);
  Left := Read(Result[1], MaxLen);
  if Left < MaxLen then
    SetLength(Result, Left)
  ;
end;

function TDFileStream.ReadRIFFChunk(var Typ: LongWord): String;
var
  Left: Integer;
  ClaimedSize, ActualSize: Integer;
  TypStr: String;

begin
  Left := Size - Position;
  if Left < 8 then raise EReadError.Create('Insufficient data to read a RIFF chunk header.');

  SetString(Result, nil, 8);
  ActualSize := Read(Result[1], 8);
  if ActualSize <> 8 then raise EReadError.Create('Buffer underrun reading RIFF chunk header.');
  Dec(Left, 8);
  TypStr := Copy(Result, 1, 4);
  Result := Copy(Result, 5, 4);
  Typ := PLongWord(PChar(TypStr))^;
  ClaimedSize := PLongWord(PChar(Result))^;

  if Left < ClaimedSize then raise EReadError.Create('Insufficient data to read content for RIFF chunk of type [' + TypStr + '].');

  SetString(Result, nil, ClaimedSize);
  ActualSize := Read(Result[1], ClaimedSize);
  if ActualSize <> ClaimedSize then raise EReadError.Create('Buffer underrun reading content for RIFF chunk of type [' + TypStr + '].');
end;

function TDFileStream.WriteStr(S: String): LongInt;
begin
  Result := Write(S[1], Length(S));
end;

end.
